import unittest
from unittest.mock import Mock

from data_repo_client import RepositoryApi
from dagster_utils.contrib.data_repo.jobs import poll_job, JobFailureException, JobTimeoutException
from dagster_utils.contrib.data_repo.typing import JobId


class PollJobTestCase(unittest.TestCase):
    def setUp(self):
        self.data_repo_client = Mock(spec=RepositoryApi)

    def test_returns_success_on_job_complete(self):
        result = poll_job(
            JobId("fake_job_id"),
            2,
            1,
            self.data_repo_client
        )

        self.assertEqual(result, "fake_job_id")

    def test_raises_on_poll_timeout(self):
        job_status_result = Mock()
        job_status_result.completed = False
        self.data_repo_client.retrieve_job = Mock(return_value=job_status_result)

        with self.assertRaises(JobTimeoutException):
            result = poll_job(
                JobId("fake_job_id"),
                2,
                1,
                self.data_repo_client
            )

            self.assertEqual(result, "fake_job_id")

    def test_raises_on_job_failure(self):
        job_status_result = Mock()
        job_status_result.completed = True
        job_status_result.job_status = 'failed'
        self.data_repo_client.retrieve_job = Mock(return_value=job_status_result)

        with self.assertRaises(JobFailureException):
            result = poll_job(
                JobId("fake_job_id"),
                2,
                1,
                self.data_repo_client
            )

            self.assertEqual(result, "fake_job_id")



from typing import Optional
from dataclasses import dataclass
from urllib.parse import urlparse

import google.auth
from google.auth.transport.requests import AuthorizedSession, Request
from google.cloud.storage.client import Client
from google.oauth2.credentials import Credentials

DEFAULT_SCOPES = ['openid', 'email', 'profile', 'https://www.googleapis.com/auth/cloud-platform']


def google_default() -> tuple[Credentials, str]:
    return google.auth.default(scopes=DEFAULT_SCOPES)  # type: ignore # (unannotated library)


def get_credentials() -> Credentials:
    creds, _ = google_default()
    return creds


def default_google_access_token(credentials: Optional[Credentials] = None) -> str:
    """
    Get token for Google-based auth use.
    Assumes application default credentials work for specified environment.
    Note that this token will only be valid for 60 minutes. For uses that may need to
    make requests for longer than this, use an AuthorizedSession (which auto-refreshes the token).
    """
    credentials = credentials or get_credentials()
    credentials.refresh(Request())

    return credentials.token  # type: ignore # (unannotated library)


def authorized_session() -> AuthorizedSession:
    return AuthorizedSession(get_credentials())


def gs_path_from_bucket_prefix(bucket: str, prefix: str) -> str:
    return f"gs://{bucket}/{prefix}"


def path_has_any_data(bucket: str, prefix: str, gcs: Client) -> bool:
    """Checks the given path for any blobs of non-zero size"""
    blobs = [blob for blob in
             gcs.list_blobs(bucket, prefix=prefix)]
    return any([blob.size > 0 for blob in blobs])


@dataclass
class GsBucketWithPrefix:
    bucket: str
    prefix: str

    def to_gs_path(self) -> str:
        return f"gs://{self.bucket}/{self.prefix}"


def parse_gs_path(raw_gs_path: str) -> GsBucketWithPrefix:
    if not raw_gs_path.startswith("gs://"):
        raise ValueError("GS path must being with gs:// scheme")
    url_result = urlparse(raw_gs_path)
    return GsBucketWithPrefix(url_result.netloc, url_result.path[1:])



import subprocess
from dataclasses import dataclass
from typing import Any, Optional

from dagster import DagsterLogManager, resource, Field, StringSource
from dagster.core.execution.context.init import InitResourceContext

from dagster_utils.resources.beam.beam_runner import BeamRunner


@dataclass
class LocalBeamRunner(BeamRunner):
    working_dir: str
    logger: DagsterLogManager

    def __post_init__(self) -> None:
        self.arg_dict = {
            "runner": "direct",
        }

    def run(
            self,
            run_arg_dict: dict[str, Any],
            target_class: str,
            scala_project: str,
            job_name: Optional[str] = None,
            command: Optional[list[str]] = None
    ) -> None:
        # create a new dictionary containing the keys and values of arg_dict + solid arguments
        local_run_flags = {**self.arg_dict, **run_arg_dict}
        self.logger.info("Local beam runner")

        # list comprehension over args_dict to get flags
        flags = " ".join([f'--{arg}={value}' for arg, value in local_run_flags.items()])
        subprocess.run(
            ["sbt", f'{scala_project}/runMain {target_class} {flags}'],
            check=True,
            cwd=self.working_dir
        )


@resource({
    "working_dir": Field(StringSource),
})
def local_beam_runner(init_context: InitResourceContext) -> LocalBeamRunner:
    return LocalBeamRunner(
        working_dir=init_context.resource_config["working_dir"],
        logger=init_context.log_manager
    )



"""
Types and type annotations useful for Dagster applications.
"""

from dagster import HookContext, InputDefinition
from dagster.config import ConfigType as DagsterConfigType

from typing import Callable, Literal, Protocol, TypedDict, Union


# dict of config settings for a given instance of a dagster object
# assumes the config is flat (i.e. only primitive types)
DagsterConfigDict = dict[
    str,
    Union[
        dict[
            Literal['env'],
            str
        ],
        str,
        int,
        float,
        bool
    ]
]

# dict representing how a solid can be configured
DagsterObjectConfigSchema = dict[str, DagsterConfigType]

DagsterHookFunction = Callable[[HookContext], None]


# a partial delineation of the config for a Dagster solid.
class DagsterSolidConfig(TypedDict, total=False):
    required_resource_keys: set[str]
    input_defs: list[InputDefinition]
    config_schema: DagsterObjectConfigSchema


# a package whose location on the filesystem is fetchable
class LocatablePackage(Protocol):
    @property
    def __file__(self) -> str: ...



from urllib.parse import urljoin
from dataclasses import dataclass

from dagster import Field, resource, StringSource
from dagster.core.execution.context.init import InitResourceContext
from google.auth.transport.requests import AuthorizedSession

from dagster_utils.contrib.google import authorized_session


@dataclass
class Sam:
    base_url: str

    def set_public_flag(self, snapshot_id: str, status: bool) -> None:
        # we are explicitly set content-type in this PUT as the requests lib only sets it when
        # using the json= kwarg, and SAM will 415 otherwise
        response = self._session().put(
            self._api_url(f'api/resources/v1/datasnapshot/{snapshot_id}/policies/reader/public'),
            headers={"Content-type": "application/json"},
            data=f"{str(status).lower()}",  # telling the endpoint to set the flag to true/false
        )

        # raise an exception for a bad response
        response.raise_for_status()

    # private methods

    def _api_url(self, url_part: str) -> str:
        return urljoin(self.base_url, url_part)

    def _session(self) -> AuthorizedSession:
        return authorized_session()


@resource({
    "api_url": Field(StringSource)
})
def sam_client(init_context: InitResourceContext) -> Sam:
    return Sam(base_url=init_context.resource_config['api_url'])


class NoopSamClient:
    def set_public_flag(self, snapshot_id: str, status: bool) -> None:
        pass


@resource
def noop_sam_client(init_context: InitResourceContext) -> NoopSamClient:
    return NoopSamClient()



import unittest

from unittest.mock import MagicMock, patch, call
from typing import Optional

from dagster_utils.contrib.argo_workflows import ArgoArchivedWorkflowsClient
from argo.workflows.client.models import V1alpha1Workflow

from dagster_utils.tests.support.mock_workflows import mock_argo_workflow, extend_workflow


class ArgoArchivedWorkflowsClientTestCase(unittest.TestCase):
    def setUp(self):
        self.client = ArgoArchivedWorkflowsClient("https://zombo.com", "tokentokentoken")

    def workflows_to_page_of_results(
            self, archived_workflows: list[V1alpha1Workflow], next_page: Optional[str]) -> MagicMock:
        results_page = MagicMock()
        results_page.items = archived_workflows
        results_page.metadata._continue = next_page
        return results_page

    def test__pull_paginated_results_api_function_no_continue_results_one_call(self):
        workflows_on_page_one = [
            mock_argo_workflow('import-hca-total-defg', 'abc123uid', 'Succeeded'),
            mock_argo_workflow('import-hca-total-abcd', 'abc234uid', 'Supsneeded', params={
                'data-repo-name': 'datarepo_dataset1'
            }),
            mock_argo_workflow('import-hca-total-cdef', 'abc345uid', 'Succeeded', params={
                'data-repo-name': 'datarepo_dataset2'
            }),
        ]

        archived_workflow_list = self.workflows_to_page_of_results(workflows_on_page_one, None)

        with patch('argo.workflows.client.ArchivedWorkflowServiceApi.list_archived_workflows',
                   return_value=archived_workflow_list) as mock_list_archived_workflows:
            list_results = list(self.client._pull_paginated_results(mock_list_archived_workflows))
            self.assertEqual(list_results, workflows_on_page_one)
            mock_list_archived_workflows.assert_called_once()

    def test__pull_paginated_results_api_function_one_continue_results_two_calls(self):
        workflows_on_page_one = [
            mock_argo_workflow('import-hca-total-defg', 'abc123uid', 'Succeeded'),
            mock_argo_workflow('import-hca-total-abcd', 'abc234uid', 'Supsneeded', params={
                'data-repo-name': 'datarepo_dataset1'
            }),
            mock_argo_workflow('import-hca-total-cdef', 'abc345uid', 'Succeeded', params={
                'data-repo-name': 'datarepo_dataset2'
            }),
        ]

        page_one_response = self.workflows_to_page_of_results(workflows_on_page_one,
                                                              "next_page")

        workflows_on_page_two = [
            mock_argo_workflow('import-hca-total-page2', 'page2uid', 'Supsneeded', params={
                'data-repo-name': 'datarepo_dataset1'
            }),
        ]

        page_two_response = self.workflows_to_page_of_results(workflows_on_page_two,
                                                              None)

        def get_page(list_options_continue=None):
            if list_options_continue is None:
                return page_one_response

            if list_options_continue == "next_page":
                return page_two_response

        with patch('argo.workflows.client.ArchivedWorkflowServiceApi.list_archived_workflows',
                   side_effect=get_page) as mock_list_archived_workflows:
            list_results = list(self.client._pull_paginated_results(mock_list_archived_workflows))
            self.assertEqual(list_results, workflows_on_page_one + workflows_on_page_two)
            self.assertEqual(mock_list_archived_workflows.call_count, 2)

            mock_list_archived_workflows.assert_has_calls([call(), call(list_options_continue="next_page")])

    def test__pull_paginated_results_api_function_multiple_continues_results_multiple_calls(self):
        workflows_on_page_one = [
            mock_argo_workflow('import-hca-total-defg', 'abc123uid', 'Succeeded'),
            mock_argo_workflow('import-hca-total-abcd', 'abc234uid', 'Supsneeded', params={
                'data-repo-name': 'datarepo_dataset1'
            }),
            mock_argo_workflow('import-hca-total-cdef', 'abc345uid', 'Succeeded', params={
                'data-repo-name': 'datarepo_dataset2'
            }),
        ]

        page_one_response = self.workflows_to_page_of_results(workflows_on_page_one,
                                                              "page2")

        workflows_on_page_two = [
            mock_argo_workflow('import-hca-total-page2', 'page2uid', 'Supsneeded', params={
                'data-repo-name': 'datarepo_dataset1'
            }),
        ]

        page_two_response = self.workflows_to_page_of_results(workflows_on_page_two,
                                                              "page3")

        workflows_on_page_three = [
            mock_argo_workflow('import-hca-1', '12345', 'Succeeded', params={
                'data-name': 'dataset',
                'test': 'data'
            }),
            mock_argo_workflow('hca-data', 'u123id', 'Succeeded'),
        ]

        page_three_response = self.workflows_to_page_of_results(workflows_on_page_three,
                                                                "page4")

        workflows_on_page_four = [
            mock_argo_workflow('hca-import', 'uid', 'Succeeded', params={
                'testing_data': 'dataset'
            }),
        ]

        page_four_response = self.workflows_to_page_of_results(workflows_on_page_four,
                                                               None)

        def get_page(list_options_continue=None):
            page_response = {
                None: page_one_response,
                "page2": page_two_response,
                "page3": page_three_response,
                "page4": page_four_response
            }[list_options_continue]
            return page_response

        with patch('argo.workflows.client.ArchivedWorkflowServiceApi.list_archived_workflows',
                   side_effect=get_page) as mock_list_archived_workflows:
            list_results = list(self.client._pull_paginated_results(mock_list_archived_workflows))
            self.assertEqual(list_results,
                             workflows_on_page_one +
                             workflows_on_page_two +
                             workflows_on_page_three +
                             workflows_on_page_four)
            self.assertEqual(mock_list_archived_workflows.call_count, 4)
            mock_list_archived_workflows.assert_has_calls([call(),
                                                           call(list_options_continue="page2"),
                                                           call(list_options_continue="page3"),
                                                           call(list_options_continue="page4")])

    def test_inflate_one_workflow_one_call(self):
        mock_inflated_workflow = mock_argo_workflow('import-hca-total-abcd', 'abc234uid', 'Succeeded', params={
            'data-repo-name': 'datarepo_dataset1'
        })

        workflow = extend_workflow(mock_inflated_workflow)

        with patch('dagster_utils.contrib.argo_workflows.ArgoArchivedWorkflowsClient.get_archived_workflow',
                   return_value=mock_inflated_workflow) as mock_get_archived_workflow:
            workflow.inflate()
            mock_get_archived_workflow.assert_called_once_with(workflow.metadata.uid)
            # Checking that despite inflate called twice, mock_get_archived_workflow is only called once
            workflow.inflate()
            mock_get_archived_workflow.assert_called_once_with(workflow.metadata.uid)

    def test_params_dict_workflow_info_dict_proper_values(self):
        params_for_argo_workflow_large = {'data-repo-name': 'datarepo_dataset1',
                                          'name': 'value',
                                          'testing': 'data',
                                          '1023_3': 'e5Qss'}
        params_for_argo_workflow_small = {
            'dataset_name': 'data'
        }
        archived_workflows = [
            mock_argo_workflow(
                'import-hca-total-abcd',
                'abc234uid',
                'Succeeded',
                params=params_for_argo_workflow_large),
            mock_argo_workflow('hca_import', 'randomuid', 'Succeeded', params=params_for_argo_workflow_small),
            mock_argo_workflow('hca_import', 'randomuid', 'Succeeded'),
        ]

        list_workflows = [extend_workflow(workflow) for workflow in archived_workflows]

        self.assertEqual(list_workflows[0].params_dict(), params_for_argo_workflow_large)
        self.assertEqual(list_workflows[1].params_dict(), params_for_argo_workflow_small)
        self.assertEqual(list_workflows[2].params_dict(), {})



from contextlib import contextmanager

from typing import Any

from dagster import DagsterInstance, ResourceDefinition
from dagster.core.execution.build_resources import build_resources
from dagster_utils.typing import DagsterConfigDict


# n.b. 2021-03-22
# dagster support for testing resources in isolation is currently in active development.
# expect this section to use more robust and unchanging tooling as it becomes available over the next few months
@contextmanager
def initialize_resource(resource_def: ResourceDefinition, config: DagsterConfigDict = {}) -> Any:
    with build_resources(
        {
            'test_resource': resource_def,
        },
        DagsterInstance.ephemeral(),
        {
            'test_resource': {
                'config': config,
            }
        }
    ) as resource_context:
        yield resource_context.test_resource



import pytest

from dagster_utils.contrib.google import parse_gs_path, GsBucketWithPrefix


def test_parse_gs_path():
    raw_path = "gs://example/bucket/path"

    result = parse_gs_path(raw_path)

    assert result.bucket == "example"
    assert result.prefix == "bucket/path"


def test_parse_gs_path_rejects_non_gs_paths():
    raw_path = "bucket/path"

    with pytest.raises(ValueError):
        parse_gs_path(raw_path)


def test_maps_object_back_to_path():
    object = GsBucketWithPrefix("bucket", "path/prefix")

    result = object.to_gs_path()

    assert result == "gs://bucket/path/prefix"



import unittest
from unittest.mock import MagicMock, patch

from dagster_utils.resources.sam import sam_client, Sam
from dagster_utils.testing.resources import initialize_resource
from dagster_utils.testing.matchers import StringEndingWith
from dagster_utils.typing import DagsterConfigDict


class SamResourceTestCase(unittest.TestCase):
    def setUp(self):
        self.dummy_config: DagsterConfigDict = {'api_url': 'https://example.com'}

    def test_resource_can_be_initialized_without_extra_config(self):
        with initialize_resource(sam_client, config=self.dummy_config) as client_instance:
            self.assertIsInstance(client_instance, Sam)

    def test_make_snapshot_public_hits_expected_url(self):
        with initialize_resource(sam_client, config=self.dummy_config) as client_instance:
            mock_authorized_session = MagicMock()
            with patch('dagster_utils.resources.sam.Sam._session', return_value=mock_authorized_session):
                client_instance.set_public_flag('fake-snapshot-id', True)
                mock_authorized_session.put.assert_called_once_with(
                    StringEndingWith('datasnapshot/fake-snapshot-id/policies/reader/public'),
                    headers={"Content-type": "application/json"},
                    data="true"
                )



import unittest

from dagster_utils.resources.slack import live_slack_client, LiveSlackClient
from dagster_utils.testing.resources import initialize_resource
from dagster_utils.typing import DagsterConfigDict


class LiveSlackResourceTestCase(unittest.TestCase):
    def test_resource_can_be_initialized(self):
        test_config: DagsterConfigDict = {'channel': '#something', 'token': 'aeiou'}
        with initialize_resource(live_slack_client, config=test_config) as client_instance:
            self.assertIsInstance(client_instance, LiveSlackClient)



from typing import Any, Optional
from dataclasses import dataclass
import subprocess

from dagster import DagsterLogManager, resource, Field, IntSource, StringSource
from dagster.core.execution.context.init import InitResourceContext

from dagster_utils.resources.beam.beam_runner import BeamRunner


@dataclass
class DataflowBeamRunner(BeamRunner):
    working_dir: str
    logger: DagsterLogManager
    region: str
    worker_machine_type: str
    autoscaling_algorithm: str
    num_workers: int
    max_num_workers: int
    google_project: str

    def __post_init__(self) -> None:
        self.arg_dict = {
            "region": self.region,
            "workerMachineType": self.worker_machine_type,
            "autoscalingAlgorithm": self.autoscaling_algorithm,
            "numWorkers": str(self.num_workers),
            "maxNumWorkers": str(self.max_num_workers),
            "project": self.google_project,
            # specific to dataflow runner
            "runner": "dataflow",
            "experiments": "shuffle_mode=service",
        }

    def run(
            self,
            run_arg_dict: dict[str, Any],
            target_class: str,
            scala_project: str,
            job_name: Optional[str] = None,
            command: Optional[list[str]] = None
    ) -> None:
        # create a new dictionary containing the keys and values of arg_dict + solid arguments
        dataflow_run_flags = {**self.arg_dict, **run_arg_dict}
        self.logger.info("Dataflow beam runner")

        # list comprehension over args_dict to get flags
        flags = " ".join([f'--{arg}={value}' for arg, value in dataflow_run_flags.items()])
        subprocess.run(
            ["sbt", f'{scala_project}/runMain {target_class} {flags}'],
            check=True,
            cwd=self.working_dir
        )


@resource({
    "working_dir": Field(StringSource),
    "region": Field(StringSource),
    "worker_machine_type": Field(StringSource),
    "autoscaling_algorithm": Field(StringSource),
    "num_workers": Field(IntSource),
    "max_num_workers": Field(IntSource),
    "google_project": Field(StringSource),
})
def dataflow_beam_runner(init_context: InitResourceContext) -> DataflowBeamRunner:
    return DataflowBeamRunner(
        working_dir=init_context.resource_config["working_dir"],
        region=init_context.resource_config["region"],
        worker_machine_type=init_context.resource_config["worker_machine_type"],
        autoscaling_algorithm=init_context.resource_config["autoscaling_algorithm"],
        num_workers=init_context.resource_config["num_workers"],
        max_num_workers=init_context.resource_config["max_num_workers"],
        google_project=init_context.resource_config["google_project"],
        logger=init_context.log_manager,
    )



from dataclasses import dataclass
import os
from typing import Optional
import warnings

import yaml

from dagster_utils.typing import DagsterConfigDict


@dataclass
class PreconfigurationLoader:
    name: str
    config_file_directory: str
    optional_keys: set[str]
    required_keys: set[str]

    def file_path(self, filename: str) -> str:
        return os.path.join(self.config_file_directory, filename)

    def load_config(self, filename: str) -> Optional[DagsterConfigDict]:
        config_file_path = self.file_path(filename)
        if os.path.exists(config_file_path):
            with open(config_file_path, 'r') as config_file_io:
                return DagsterConfigDict(yaml.safe_load(config_file_io))

        return None

    # raises an error for any missing config keys, records a warning for (and discards) any extra keys
    def validated_config(self, config: DagsterConfigDict) -> DagsterConfigDict:
        keys_present = set(config.keys())
        permitted_keys = self.required_keys | self.optional_keys
        missing_keys = self.required_keys - keys_present
        extra_keys = keys_present - permitted_keys

        if any(missing_keys):
            missing_keys_str = ", ".join(key for key in missing_keys)
            raise ValueError(
                f"Missing expected preconfigured field(s) in configuration files for {self.name}. "
                f"Config files do not define these required fields:\n{missing_keys_str}\n"
                f"(config dir: {self.config_file_directory})"
            )

        if any(extra_keys):
            extra_keys_str = ", ".join(key for key in extra_keys)
            warnings.warn(
                message=(
                    f"Found unexpected fields in configuration files for {self.name}. "
                    "These fields will be ignored. "
                    f"Fields:\n{extra_keys_str}\n(config dir: {self.config_file_directory})"
                )
            )

        return {k: v for k, v in config.items() if k in permitted_keys}

    # loads a list of config files, returning only those that exist
    def load_files(self, filenames: list[str]) -> list[DagsterConfigDict]:
        configs = [
            self.load_config(config_name)
            for config_name in filenames
        ]
        if not any(configs):
            expected_files_str = ', '.join(filenames)
            raise FileNotFoundError(
                f"No configuration files detected for {self.name}! "
                f"Expected at least one of these files in {self.config_file_directory}:\n{expected_files_str}"
            )

        return [config for config in configs if config is not None]

    # loads the global and mode-specific config for the given mode,
    # and verifies that the loaded config is valid.
    def load_for_mode(self, mode: str) -> DagsterConfigDict:
        configs = self.load_files(['global.yaml', f"{mode}.yaml"])

        loaded_config: DagsterConfigDict = {}

        for config in configs:
            loaded_config = {
                **loaded_config,
                **config,
            }

        return self.validated_config(loaded_config)



from google.cloud.bigquery import Dataset, Client
from dagster import resource, InitResourceContext

from dagster_utils.contrib.google import authorized_session

from unittest.mock import Mock


@resource
def bigquery_client(init_context: InitResourceContext) -> Client:
    return Client(_http=authorized_session())


@resource
def noop_bigquery_client(init_context: InitResourceContext) -> Client:
    return Mock(spec=Client)



"""
Types and annotations useful for interaction with TDR
"""


# Aliases to add some type safety when calling the jade API client and make juggling the (many) string
# parameters easier

class JobId(str):
    pass


class ProfileId(str):
    pass


class DatasetName(str):
    pass


class DatasetId(str):
    pass



from dataclasses import dataclass, field
from typing import List, Any, Optional
from uuid import uuid4

import kubernetes
from dagster import DagsterLogManager, Field, IntSource, resource, StringSource, String, Noneable, BoolSource
from dagster.core.execution.context.init import InitResourceContext
from dagster_k8s.client import DagsterKubernetesClient
from kubernetes.client.models.v1_job import V1Job

from dagster_utils.resources.beam.beam_runner import BeamRunner


# separating out config for the cloud dataflow pipeline to make
# the giant mess of parameters here a little easier to parse
@dataclass
class K8sDataflowCloudConfig:
    project: str
    service_account: str
    subnet_name: Optional[str]
    region: str
    worker_machine_type: str
    starting_workers: int
    max_workers: int
    load_incluster_config: bool

    def subnetwork(self) -> Optional[str]:
        if not self.subnet_name:
            return None

        return '/'.join([
            'regions',
            self.region,
            'subnetworks',
            self.subnet_name,
        ])


@dataclass
class K8sDataflowBeamRunner(BeamRunner):
    cloud_config: K8sDataflowCloudConfig
    kubernetes_service_account: str
    temp_bucket: str
    image_name: str
    image_version: str
    namespace: str
    logger: DagsterLogManager

    def run(
            self,
            run_arg_dict: dict[str, Any],
            target_class: str,
            scala_project: str,
            job_name: Optional[str] = None,
            command: Optional[list[str]] = None
    ) -> None:
        args_dict = {
            'runner': 'dataflow',
            'project': self.cloud_config.project,
            'region': self.cloud_config.region,
            'tempLocation': f'gs://{self.temp_bucket}/dataflow',
            'serviceAccount': self.cloud_config.service_account,
            'workerMachineType': self.cloud_config.worker_machine_type,
            'autoscalingAlgorithm': 'THROUGHPUT_BASED',
            'numWorkers': str(self.cloud_config.starting_workers),
            'maxNumWorkers': str(self.cloud_config.max_workers),
            'experiments': 'shuffle_mode=service',
        }
        subnet = self.cloud_config.subnetwork()
        if subnet:
            args_dict['subnetwork'] = subnet

        args_dict = {**args_dict, **run_arg_dict}

        # process the args dict into dataflow flags
        args = [
            f'--{field_name}={value}'
            for field_name, value
            in args_dict.items()
        ]

        image_name = f"{self.image_name}:{self.image_version}"  # {context.solid_config['version']}"
        job = self.dispatch_k8s_job(image_name, job_name, args, command=command,
                                    load_incluster_config=self.cloud_config.load_incluster_config)
        self.logger.info("Dataflow job started")

        client = DagsterKubernetesClient.production_client()
        client.wait_for_job_success(job.metadata.name, self.namespace)

    def dispatch_k8s_job(
            self,
            image_name: str,
            job_name_prefix: Optional[str],
            args: List[str],
            load_incluster_config: bool = True,
            command: Optional[list[str]] = None
    ) -> V1Job:
        # we will need to poll the pod/job status on creation
        if load_incluster_config:
            kubernetes.config.load_incluster_config()
        else:
            kubernetes.config.load_kube_config()

        if job_name_prefix:
            job_name = f"{job_name_prefix}-{uuid4()}"
        else:
            job_name = f"{uuid4()}"

        pod_name = f"{job_name}-pod"
        job_container = kubernetes.client.V1Container(
            name=job_name,
            image=image_name,
            args=args,
            command=command
        )

        template = kubernetes.client.V1PodTemplateSpec(
            metadata=kubernetes.client.V1ObjectMeta(name=pod_name),
            spec=kubernetes.client.V1PodSpec(
                restart_policy="Never",
                containers=[job_container],
                service_account_name=self.kubernetes_service_account,
            ),
        )

        job = kubernetes.client.V1Job(
            api_version="batch/v1",
            kind="Job",
            metadata=kubernetes.client.V1ObjectMeta(
                name=job_name,
            ),
            spec=kubernetes.client.V1JobSpec(
                template=template,
                backoff_limit=2,
                ttl_seconds_after_finished=86400,  # one day
            ),
        )
        batch_v1 = kubernetes.client.BatchV1Api()
        api_response = batch_v1.create_namespaced_job(
            body=job,
            namespace=self.namespace,
        )
        self.logger.info(f"Job created. status='{str(api_response.status)}'")

        return api_response


@resource({
    "project": Field(StringSource),
    "subnet_name": Field(Noneable(str)),
    "region": Field(StringSource),
    "worker_machine_type": Field(StringSource),
    "starting_workers": Field(IntSource),
    "max_workers": Field(IntSource),
    "service_account": Field(StringSource),
    "kubernetes_service_account": Field(StringSource),
    "temp_bucket": Field(StringSource),
    "image_name": Field(StringSource),
    "image_version": Field(StringSource),
    "namespace": Field(StringSource),
    "load_incluster_config": Field(BoolSource)
})
def k8s_dataflow_beam_runner(init_context: InitResourceContext) -> K8sDataflowBeamRunner:
    cloud_config = K8sDataflowCloudConfig(
        project=init_context.resource_config['project'],
        service_account=init_context.resource_config['service_account'],
        subnet_name=init_context.resource_config['subnet_name'],
        region=init_context.resource_config['region'],
        worker_machine_type=init_context.resource_config['worker_machine_type'],
        starting_workers=init_context.resource_config['starting_workers'],
        max_workers=init_context.resource_config['max_workers'],
        load_incluster_config=init_context.resource_config["load_incluster_config"]
    )
    return K8sDataflowBeamRunner(
        cloud_config=cloud_config,
        kubernetes_service_account=init_context.resource_config['kubernetes_service_account'],
        temp_bucket=init_context.resource_config['temp_bucket'],
        image_name=init_context.resource_config['image_name'],
        image_version=init_context.resource_config['image_version'],
        namespace=init_context.resource_config['namespace'],
        logger=init_context.log_manager
    )



import unittest
import pytest
from dagster_utils.resources.data_repo.jade_data_repo import build_client


class DataRepoClientTestCase(unittest.TestCase):

    @pytest.mark.ci_only
    def test_client_auths_successfully(self):
        # make sure we can successfully connect to Jade
        # this test is marked ci-only as there is some gcloud dependent configuration
        # needed which may be cumbersome in local dev
        client = build_client(host='https://jade.datarepo-dev.broadinstitute.org/')
        result = client.enumerate_datasets()

        self.assertIsNotNone(result, "Should get back any result")



from typing import Iterator
from dataclasses import dataclass

from dagster import resource
from dagster.core.execution.context.init import InitResourceContext

from google.cloud import storage

from dagster_utils.contrib.google import authorized_session, google_default


@resource
def google_storage_client(_: InitResourceContext) -> storage.Client:
    _, project = google_default()

    return storage.Client(project=project, _http=authorized_session())


@dataclass
class MockBlob:
    name: str
    size: int

    def delete(self) -> None:
        pass


class MockStorageClient:
    def list_blobs(self, bucket_name: str, prefix: str) -> Iterator[MockBlob]:
        for i in range(0, 10):
            yield MockBlob(f"fake_blob_{i}", 1)


@resource
def mock_storage_client(_: InitResourceContext) -> MockStorageClient:
    return MockStorageClient()



import os
from typing import Optional, Protocol

from dagster import configured, Noneable
from dagster.core.definitions.configurable import ConfigurableDefinition

from dagster_utils.typing import DagsterConfigDict, DagsterObjectConfigSchema, LocatablePackage
from dagster_utils.config.preconfiguration_loader import PreconfigurationLoader


class PreconfiguratorFunction(Protocol):
    def __call__(
        self,
        dagster_object: ConfigurableDefinition,
        mode_name: str,
        additional_schema: DagsterObjectConfigSchema = {},
        subdirectory: Optional[str] = None,
    ) -> ConfigurableDefinition: ...


def configurator_aimed_at(
    package: LocatablePackage,
) -> PreconfiguratorFunction:
    return configurator_aimed_at_directory(os.path.dirname(package.__file__))


def configurator_aimed_at_directory(
    package_directory: str,
) -> PreconfiguratorFunction:
    """
    Generates a configuration function that loads config from YAML within the specified package directory.
    See the inner function for details on how it works.

    Example usage:
    preconfigure_for_mode = configurator_aimed_at(os.path.dirname(my_dagster_project.config.__file__))

    preconfigure_for_mode(some_resource, "dev") # => configuration pulled from [package_root]/my_dagster_project/config/some_resource/
    """

    def preconfigure_for_mode(
        dagster_object: ConfigurableDefinition,
        mode_name: str,
        additional_schema: DagsterObjectConfigSchema = {},
        subdirectory: Optional[str] = None,
    ) -> ConfigurableDefinition:
        """
        Preconfigures a Dagster object (such as a resource) for a given mode by setting all config values
        for the object to values found in the config files in the specified directory.

        The method will load from the specified directory under [base_config_package_dir]/config/, looking for:
        * global.yaml
        * [mode_name].yaml

        Values in the mode-specific config wil supersede global config. Only one of these needs to be present.

        This method expects a value to be specified for all required fields in the object's config schema. If
        a value should remain configurable after preconfiguration, list it in the `additional_schema` argument.
        If any values it expects to be specified are missing, it will raise an error. If any values for
        fields it does not expect are specified, it will record a warning and ignore those values.

        If a config field is optional (i.e. it has the Noneable type), it does not need to be configured in the YAML,
        but it will be unconfigurable in the resulting solid unless it is listed in additional_schema.

        :param dagster_object: The definition for the object to be configured (e.g. a ResourceDefinition).
        :param mode_name: The name of the mode. This will determine the name of the mode-specific config file to load.
        :param additional_schema: Any config schema that is part of the Dagster object but should NOT be preconfigured.
        :param subdirectory: Which directory to search for the config files. Defaults to the name of the Dagster object
            beneath your_repo_name_here.config.
        :return: The Dagster object configured with the loaded values.
        """

        # [NOTE]
        # This line assumes the Dagster object is configured with a dict. Other dagster config patterns,
        # such as a single primitive type, aren't accounted for (but probably shouldn't be used,
        # since they provide no context for the config setting's meaning/purpose)
        definition_config_keys = dagster_object.config_schema.config_type.fields
        optional_config_keys = [k for k, v in definition_config_keys.items() if isinstance(v.config_type, Noneable)]
        required_config_keys = [k for k, v in definition_config_keys.items() if k not in optional_config_keys]
        subdirectory = subdirectory or dagster_object.__name__
        config_file_directory = os.path.join(package_directory, subdirectory)
        loader = PreconfigurationLoader(
            name=dagster_object.__name__,
            config_file_directory=config_file_directory,
            required_keys=(set(required_config_keys) - set(additional_schema.keys())),
            optional_keys=(set(optional_config_keys) - set(additional_schema.keys()))
        )

        # we load the config in preconfigure_for_mode instead of in the @configured function to
        # ensure that any config issues cause errors upon initialization, instead of waiting until
        # we try to use the object being configured
        loaded_config = loader.load_for_mode(mode_name)

        @configured(dagster_object, additional_schema)
        def __dagster_object_preconfigured(extra_config: DagsterConfigDict) -> DagsterConfigDict:
            return {
                **loaded_config,
                **extra_config,
            }

        return __dagster_object_preconfigured

    return preconfigure_for_mode



import unittest
from unittest.mock import Mock

from dagster_utils.resources.data_repo.api_configuration import RefreshingAccessTokenConfig
from dagster_utils.contrib.google import Credentials


class RefreshingConfigurationTestCase(unittest.TestCase):
    def test_refreshing_configuration_delegates_access_token(self):
        creds = Mock(spec=Credentials)
        creds.token = "faketoken!"
        creds.valid = True

        config = RefreshingAccessTokenConfig("http://example.com", creds)

        self.assertEqual(config.access_token, "faketoken!")
        # valid creds should not require refresh
        creds.refresh.assert_not_called

    def test_refreshing_configuration_refreshes_on_invalid_token(self):
        creds = Mock(spec=Credentials)
        creds.token = "faketoken!"
        creds.valid = False

        config = RefreshingAccessTokenConfig("http://example.com", creds)

        self.assertEqual(config.access_token, "faketoken!")
        # invalid creds should refresh
        creds.refresh.assert_called_once



from typing import Any, Optional

from dagster import resource
from dagster.core.execution.context.init import InitResourceContext

from dagster_utils.resources.beam.beam_runner import BeamRunner


class NoopBeamBeamer(BeamRunner):
    def run(
            self,
            run_arg_dict: dict[str, Any],
            target_class: str,
            scala_project: str,
            job_name: Optional[str] = None,
            command: Optional[list[str]] = None
    ) -> None:
        pass


@resource
def noop_beam_runner(init_context: InitResourceContext) -> NoopBeamBeamer:
    return NoopBeamBeamer()



import os
from tempfile import TemporaryDirectory
import unittest
import yaml

from dagster import solid, Noneable, String

import dagster_utils.config
from dagster_utils.config import configurator_aimed_at
from dagster_utils.testing.filesystem import EphemeralNamedDirectory


class ConfiguratorAimedAtTestCase(unittest.TestCase):
    def setUp(self):
        self.dummy_function = lambda _: 0
        self.dummy_function.__name__ = 'steve'
        self.config_package_dir = os.path.dirname(dagster_utils.config.__file__)
        self.configurator = configurator_aimed_at(dagster_utils.config)

    def test_treats_noneable_fields_as_optional(self):
        solid_def = solid(config_schema={
            'a': Noneable(String),
            'b': String,
        })(self.dummy_function)

        with TemporaryDirectory(dir=self.config_package_dir) as temp_dir:
            with open(os.path.join(temp_dir, 'mode.yaml'), 'w') as config_yaml_io:
                yaml.dump({'b': 'steve'}, config_yaml_io)

            preconfigured = self.configurator(
                solid_def,
                'mode',
                subdirectory=os.path.basename(temp_dir))
            # assert that it accepts no additional config
            self.assertEqual(preconfigured.get_config_field().config_type.fields, {})

    def test_marks_non_noneable_fields_as_required(self):
        solid_def = solid(config_schema={
            'a': String,
            'b': String,
        })(self.dummy_function)

        with TemporaryDirectory(dir=self.config_package_dir) as temp_dir:
            with open(os.path.join(temp_dir, 'mode.yaml'), 'w') as config_yaml_io:
                yaml.dump({'b': 'steve'}, config_yaml_io)

            with self.assertRaises(ValueError):
                self.configurator(
                    solid_def,
                    'mode',
                    subdirectory=os.path.basename(temp_dir))

    def test_defaults_to_using_resource_name_for_directory(self):
        solid_def = solid(config_schema={
            'a': String,
            'b': String,
        })(self.dummy_function)

        with EphemeralNamedDirectory('steve', self.config_package_dir) as temp_dir:
            with open(os.path.join(temp_dir, 'global.yaml'), 'w') as config_yaml_io:
                yaml.dump({'a': 'sneve', 'b': 'steve'}, config_yaml_io)

            preconfigured = self.configurator(solid_def, 'mode')
            # assert that it accepts no additional config
            self.assertEqual(preconfigured.get_config_field().config_type.fields, {})

    def test_doesnt_require_fields_in_additional_schema(self):
        solid_def = solid(config_schema={
            'a': String,
            'b': String,
        })(self.dummy_function)

        with EphemeralNamedDirectory('steve', self.config_package_dir) as temp_dir:
            with open(os.path.join(temp_dir, 'global.yaml'), 'w') as config_yaml_io:
                yaml.dump({'a': 'sneve'}, config_yaml_io)

            preconfigured = self.configurator(solid_def, 'mode', additional_schema={'b': String})
            # assert that it accepts no additional config
            fields = preconfigured.get_config_field().config_type.fields
            self.assertEqual(set(fields.keys()), {'b'})



"""
Common TDR ingestion functions.
"""
import logging

from data_repo_client.api.repository_api import RepositoryApi
from google.cloud.bigquery.client import Client

from dagster_utils.contrib.data_repo.typing import JobId, DatasetId, DatasetName


def ingest_tabular_json_data(
        path: str,
        table: str,
        target_dataset_id: DatasetId,
        data_repo_client: RepositoryApi,
) -> JobId:
    """
    Given a GS path with tabular JSON data, submits the data for ingestion to TDR
    :param path: A GS path; makes no assertions about the state of permissions on the path
    :param table: Target table in TDR for the data
    :param target_dataset_id: Target dataset ID of the data
    :param data_repo_client: Datarepo client appropriate for the target dataset's environment (dev, prod, etc.)
    :return: Job ID for the TDR ingestion job
    """
    if not path.startswith("gs://"):
        raise ValueError("Path must be a gs:// url")

    ingest_payload = {
        "format": "json",
        "ignore_unknown_values": "false",
        "max_bad_records": 0,
        "path": path,
        "table": table,
    }

    job_response = data_repo_client.ingest_dataset(
        id=target_dataset_id, ingest=ingest_payload
    )
    job_id = JobId(job_response.id)
    logging.info(f"Ingest job for table {table} submitted, job_id = {job_id}")

    return job_id


def find_outdated_rows_csv(
        target_dataset_project_id: str,
        target_dataset_name: DatasetName,
        table_name: str,
        primary_keys: set[str],
        version_key: str,
        outdated_ids_path: str,
        bigquery_client: Client
) -> None:
    """
    Determines rows that are outdated in a given table, assuming a monotonically increasing versioning key.
    This function assumes rows are outdated when there is a duplicate row with the same primary key and a strictly
    _higher_ value for it's version field

    :param target_dataset_project_id: Target BQ project ID
    :param target_dataset_name: Target TDR dataset name
    :param table_name: Table name to inspect for outdated rows
    :param primary_keys: Set of keys that uniquely identify a row
    :param version_key: Field on the table that should always be _higher_ for semantically newer data
    :param outdated_ids_path: Path which will receive the CSV of outdated rows
    :param bigquery_client:  BQ client
    """
    if not primary_keys:
        raise ValueError("primary_keys should be non-empty")

    if not outdated_ids_path.startswith("gs://"):
        raise ValueError("outdated_ids_path should be a gs URL")

    fully_qualified_table_name = f"{target_dataset_project_id}.datarepo_{target_dataset_name}.{table_name}"
    joined_primary_keys = ", ".join(primary_keys)
    join_points = [
        f"j.{key} = s.{key}" for key in primary_keys
    ]
    join_clause = " AND ".join(join_points)

    query = f"""
        EXPORT DATA OPTIONS(
            uri='{outdated_ids_path}',
            format='CSV',
            overwrite=true
        ) AS
        WITH latest_versions AS (
            SELECT {joined_primary_keys}, max({version_key}) as latest_version FROM `{fully_qualified_table_name}`
            GROUP BY {joined_primary_keys}
        )
        SELECT s.datarepo_row_id FROM `{fully_qualified_table_name}` s
        INNER JOIN latest_versions j ON {join_clause} WHERE s.{version_key} < j.latest_version
    """

    bigquery_client.query(query).result()


def submit_soft_deletes_csv(path: str, target_dataset_id: DatasetId, data_repo_client: RepositoryApi) -> JobId:
    """
    Submits a path containing CSV(s) of TDR row IDs for soft deletion
    :param path: GS path containing the CSV(s)
    :param target_dataset_id: TDR target dataset
    :param data_repo_client: Datarepo client appropriate for the target dataset's environment (dev, prod, etc.)
    :return: Job ID for the TDR soft deletion job
    """
    if not path.startswith("gs://"):
        raise ValueError("path should be a gs url")

    payload = {
        "deleteType": "soft",
        "specType": "gcsFile",
        "tables": [
            {
                "gcsFileSpec": {"fileType": "csv", "path": path},
                "tableName": "weather_data",
            }
        ],
    }

    job_response = data_repo_client.apply_dataset_data_deletion(
        id=target_dataset_id, data_deletion_request=payload
    )

    job_id = JobId(job_response.id)
    return job_id



# Classes with more permissive __eq__ functions for use as argument matchers for mocks.

import re
from typing import Any


# Matches if a mock was called with a string containing the given substring
class StringContaining(str):
    def __eq__(self, other_str: object) -> bool:
        return isinstance(other_str, str) and self in other_str

    def __repr__(self) -> str:
        return f"(string containing '{super().__str__()}')"


class StringStartingWith(str):
    def __eq__(self, other_str: object) -> bool:
        return isinstance(other_str, str) and other_str.startswith(self)

    def __repr__(self) -> str:
        return f"(string starting with '{super().__str__()}')"


class StringEndingWith(str):
    def __eq__(self, other_str: object) -> bool:
        return isinstance(other_str, str) and other_str.endswith(self)

    def __repr__(self) -> str:
        return f"(string ending with '{super().__str__()}')"


# doesn't currently accept compiled regexes
class StringMatchingRegex(str):
    def __eq__(self, other_str: object) -> bool:
        return isinstance(other_str, str) and re.match(self, other_str) is not None

    def __repr__(self) -> str:
        return f"(string matching regex '{super().__str__()}')"

    def __hash__(self) -> int:
        return super().__hash__()


# matches any instance of the specified type (or a descendant)
class ObjectOfType:
    def __init__(self, expected_type: type):
        self.expected_type = expected_type

    def __eq__(self, other_obj: object) -> bool:
        return isinstance(other_obj, self.expected_type)

    def __repr__(self) -> str:
        return f"(any {self.expected_type.__name__})"


# lets us do partial matching on objects without needing to make assertions about every attribute,
# e.g. "any Book with title='abc'": ObjectWithAttributes(Book, title='abc')
class ObjectWithAttributes(ObjectOfType):
    def __init__(self, expected_type: type, **expected_attributes: Any):
        super().__init__(expected_type)
        self.attributes_dict = expected_attributes

    def __eq__(self, other_obj: object) -> bool:
        return super().__eq__(other_obj) and all(
            hasattr(other_obj, attr_name) and getattr(other_obj, attr_name) == attr_value
            for attr_name, attr_value in self.attributes_dict.items()
        )

    def __repr__(self) -> str:
        attr_assertions = [
            f"{attr_name}={repr(attr_value)}"
            for attr_name, attr_value in self.attributes_dict.items()
        ]
        return f"(any {self.expected_type.__name__} with {', '.join(attr_assertions)})"



from dataclasses import dataclass
from typing import Callable, Literal, Optional

import slack
from dagster import DagsterLogManager, failure_hook, HookContext, HookDefinition,\
    resource, String, StringSource, success_hook
from dagster.core.execution.context.init import InitResourceContext

from dagster_utils.typing import DagsterHookFunction

SlackMessageGenerator = Callable[[HookContext], str]


@dataclass
class ConsoleSlackClient:
    logger: DagsterLogManager

    def send_message(self, text: Optional[str] = None, blocks: Optional[list[dict[str, object]]] = None) -> None:
        self.logger.info(f"[SLACK] {text} {blocks}")


@resource
def console_slack_client(init_context: InitResourceContext) -> ConsoleSlackClient:
    return ConsoleSlackClient(init_context.log)


@dataclass
class LiveSlackClient:
    client: slack.WebClient
    channel: str

    def send_message(self, text: Optional[str] = None, blocks: Optional[list[dict[str, object]]] = None) -> None:
        self.client.chat_postMessage(channel=self.channel, text=text, blocks=blocks)


@resource({
    'channel': String,
    'token': StringSource,
})
def live_slack_client(init_context: InitResourceContext) -> LiveSlackClient:
    return LiveSlackClient(
        slack.WebClient(init_context.resource_config['token']),
        init_context.resource_config['channel'],
    )


def slack_hook(
    on: Literal['success', 'failure'],
    additional_resource_keys: set[str] = set()
) -> Callable[[SlackMessageGenerator], DagsterHookFunction]:
    """
    Decorator function for Slack hooks - decorate a function that generates a message to turn it
    into a boilerplate Slack hook you can bind to solids/pipelines. Presumes that the Slack resource
    is already fully configured with its target channel and token.

    Example usage (real world example):
        @slack_hook(on='success', additional_resource_keys={'ornithologist'})
        def do_owls_exist(context: HookContext) -> str:
            if context.resources.ornithologist.are_there_owls():
                return "heck yeah"

            return "no owls in this world"

        # then you can call:
        create_world_with_potential_for_owls.with_hooks({do_owls_exist})()
    """

    if on == 'success':
        hook_decorator = success_hook
    elif on == 'failure':
        hook_decorator = failure_hook

    def message_generation_wrapper(message_generator: SlackMessageGenerator) -> HookDefinition:

        @hook_decorator(required_resource_keys=({'slack'} | additional_resource_keys))
        def _send_slack_message(context: HookContext) -> None:
            context.resources.slack.send_message(message_generator(context))

        return _send_slack_message

    return message_generation_wrapper



from typing import Any, Protocol, Optional


class BeamRunner(Protocol):
    def run(
            self,
            run_arg_dict: dict[str, Any],
            target_class: str,
            scala_project: str,
            job_name: Optional[str] = None,
            command: Optional[list[str]] = None
    ) -> None:
        ...



from datetime import datetime
from dateutil.tz import tzlocal
from typing import Any

from argo.workflows.client.models import V1alpha1Workflow, V1ObjectMeta, V1alpha1WorkflowSpec, V1alpha1Arguments, V1alpha1Parameter, \
    V1alpha1WorkflowStatus

from dagster_utils.contrib.argo_workflows import ExtendedArgoWorkflow

# the argo workflows api produces this abominable nested set of classes for each workflow,
# so we build one from simple params here to keep our tests lean


def mock_argo_workflow(name: str, uid: str, status: str, finished_at: datetime = datetime.now(
        tz=tzlocal()), params: dict[str, Any] = {}) -> V1alpha1Workflow:
    return V1alpha1Workflow(
        metadata=V1ObjectMeta(
            name=name,
            uid=uid,
        ),
        spec=V1alpha1WorkflowSpec(
            arguments=V1alpha1Arguments(
                parameters=[
                    V1alpha1Parameter(name=k, value=v)
                    for k, v in params.items()
                ]
            )
        ),
        status=V1alpha1WorkflowStatus(
            phase=status,
            finished_at=finished_at,
        )
    )


def extend_workflow(workflow: V1alpha1Workflow) -> ExtendedArgoWorkflow:
    return ExtendedArgoWorkflow(workflow, argo_url='https://nonexistentsite.test', access_token='token')



import time
from typing import Any
import logging

from data_repo_client import RepositoryApi

from dagster_utils.contrib.data_repo.typing import JobId


class JobPollException(Exception):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args)
        self.message = kwargs.get('message')


class JobFailureException(JobPollException):
    pass


class JobTimeoutException(JobPollException):
    pass


def poll_job(
        job_id: JobId,
        max_wait_time_seconds: int,
        poll_interval_seconds: int,
        data_repo_client: RepositoryApi
) -> JobId:
    """
    Polls on the given TDR job_id for max_wait_time_seconds, every poll_interval_seconds, until the job is reported
    completed or we time out.
    """
    time_waited = 0

    while time_waited < max_wait_time_seconds:
        logging.info(f"Polling on data repo job_id = {job_id}...")
        job_info = data_repo_client.retrieve_job(job_id)

        if job_info.completed:
            logging.info(f"Data repo job_id = {job_id} completed")
            if job_info.job_status == "failed":
                raise JobFailureException(
                    message="Job did not complete successfully."
                )
            return job_id

        logging.info(f"Data repo job_id = {job_id} not complete, scheduling retry...")
        time.sleep(poll_interval_seconds)
        time_waited += poll_interval_seconds

    raise JobTimeoutException(
        message=f"Exceeded max wait time of {max_wait_time_seconds} polling for status of job {job_id}."
    )



"""
TDR snapshot operations
"""
import logging
from enum import Enum

from data_repo_client import RepositoryApi, SnapshotRequestModel, SnapshotRequestContentsModel

from dagster_utils.contrib.data_repo.typing import JobId, ProfileId, DatasetName


class SnapshotMode(Enum):
    BY_ASSET = "byAsset"
    BY_FULL_VIEW = "byFullView"
    BY_ROW_ID = "byRowId"
    BY_QUERY = "byQuery"


def submit_snapshot_request(
        snapshot_name: str,
        profile_id: ProfileId,
        dataset_name: DatasetName,
        mode: SnapshotMode,
        reader_list: list[str],
        data_repo_client: RepositoryApi) -> JobId:
    snapshot_request = SnapshotRequestModel(
        name=snapshot_name,
        profile_id=profile_id,
        description=f"Create snapshot {snapshot_name}",
        contents=[SnapshotRequestContentsModel(dataset_name=dataset_name, mode=mode.value)],
        readers=reader_list
    )

    logging.info(snapshot_request)

    response = data_repo_client.create_snapshot(
        snapshot=snapshot_request
    )

    logging.info(f"Snapshot creation job id: {response.id}")
    return JobId(response.id)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dagster_utils',
 'dagster_utils.config',
 'dagster_utils.contrib',
 'dagster_utils.contrib.data_repo',
 'dagster_utils.resources',
 'dagster_utils.resources.beam',
 'dagster_utils.resources.data_repo',
 'dagster_utils.testing',
 'dagster_utils.tests',
 'dagster_utils.tests.config',
 'dagster_utils.tests.contrib',
 'dagster_utils.tests.contrib.data_repo',
 'dagster_utils.tests.resources',
 'dagster_utils.tests.resources.data_repo',
 'dagster_utils.tests.support']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'argo-workflows>=5.0.0,<6.0.0',
 'dagster>=0.12.3,<0.13.0',
 'data-repo-client>=1.134.0,<2.0.0',
 'google-cloud-bigquery>=2.15.0,<3.0.0',
 'google-cloud-storage>=1.38.0,<2.0.0',
 'slackclient>=2.9.3,<3.0.0']

setup_kwargs = {
    'name': 'broad-dagster-utils',
    'version': '0.6.6',
    'description': 'Common utilities and objects for building Dagster pipelines',
    'long_description': '# Dagster Utils\n\nA collection of common utilities used by the Monster team to develop Dagster pipelines. Each subsection has its own readme explaining its contents.\n\n## How to use this library\n\nThis library is hosted on [PyPI](https://pypi.org/) and will work with most Python package managers out of the box. You can reference this package as `broad_dagster_utils`, e.g. for Poetry:\n\n```\n# pyproject.toml\n[tool.poetry.dependencies]\nbroad_dagster_utils = "^0.1.0"\n```\n\nNote that, despite this name, you\'ll need to import the package as `dagster_utils`, e.g.:\n```python\nfrom dagster_utils.typing import DagsterConfigDict\n```\n\n### Local Testing\nFor development against a local checkout in another project (i.e., a project with a dependency on `dagster_utils`), make the following adjustment to the project\'s `pyproject.toml`:\n```\nbroad-dagster-utils = {path = "<relative path to your dagster_utils checkout>", develop = true}\n```\n\n## Versioning\n\nThis library is versioned semantically.\n\n* **Major versions** (e.g. 1.4.6 -> 2.0.0) represent significant shifts in functionality and may alter the call signatures of core features of the library. You may need to follow a migration plan to upgrade to a new major version.\n* **Minor versions** (e.g. 1.4.6 -> 1.5.0) represent the removal or alteration of specific library features. They will not change core functionality, but the changelog should be reviewed before upgrading to avoid unexpected feature removals.\n* **Patch versions** (e.g. 1.4.6 -> 1.4.7) represent internal improvements or new features that do not alter existing functionality or call signatures. They may introduce deprecations, but they will never remove deprecated functions. You can always safely upgrade to a new patch version.\n* **Prerelease versions** (e.g. 1.4.6 -> 1.4.7-alpha.1) represent changes that have not yet been made part of an official release. See "Releasing a new version" below for more info.\n\n### Describing changes\n\nWhen describing changes made in a commit message, we want to be more thorough than usual, since bugs in dependencies are harder to diagnose. Break down the changes into these categories (omitting any categories that don\'t apply):\n\n* **New features** are changes that add functionality, such as new optional arguments for existing functions or entire new classes/functions. These changes should never require that existing code be changed.\n* **Bugfixes** are fairly self-explanatory - bugs that were identified and addressed. If fixing a bug required removing or altering existing features, make sure to list those under "breaking changes" as well.\n* **Deprecations** are features that are still usable, but have been marked as deprecated (and thus trigger a warning when used). They are planned to be removed in a future version. Always try to deprecate functionality before it\'s removed.\n* **Breaking changes** are changes that may break existing code using this library, such as renaming, removing, or reordering arguments to a function, deleting functionality (including deprecated functionality), or otherwise altering the library in ways that users will need to account for. Users should be able to use this section as a complete guide to upgrading their applications to be compatible with the new version.\n\n### Releasing a new version\n\nTo release a new version, determine what type of version increase your changes constitute (see the above guide) and update the version listed in `pyproject.toml` accordingly. Poetry has several [version bump commands](https://python-poetry.org/docs/cli/#version) to help with this. You can update the version in a dedicated PR or as part of another change. When a PR that updates the version number lands on master, an action will run to create a new tag for that version number, followed by cutting a Git release and publishing the new version to PyPI.\n',
    'author': 'Monster Dev',
    'author_email': 'monsterdev@broadinstitute.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/broadinstitute/dagster-utils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import os
from tempfile import TemporaryDirectory
import unittest
import warnings
import yaml


import dagster_utils.config
from dagster_utils.config.preconfiguration_loader import PreconfigurationLoader


class PreconfigurationLoaderTestCase(unittest.TestCase):
    def setUp(self):
        self.config_package_dir = os.path.dirname(dagster_utils.config.__file__)

    def test_validated_config_raises_if_missing_keys(self):
        loader = PreconfigurationLoader(
            name="foo",
            config_file_directory="bar/baz",
            optional_keys=set(),
            required_keys={'a', 'b', 'c'}
        )

        with self.assertRaisesRegex(ValueError, 'Missing expected preconfigured field'):
            loader.validated_config({'a': 'value'})

    def test_validated_config_does_not_raise_if_missing_optional_keys(self):
        loader = PreconfigurationLoader(
            name="foo",
            config_file_directory="bar/baz",
            optional_keys={'b', 'c'},
            required_keys={'a'},
        )
        self.assertEqual(
            loader.validated_config({'a': 'value', 'b': 'thing'}),
            {'a': 'value', 'b': 'thing'}
        )

    def test_validated_config_warns_and_cuts_if_extra_keys(self):
        loader = PreconfigurationLoader(
            name="foo",
            config_file_directory="bar/baz",
            optional_keys=set(),
            required_keys={'a'},
        )

        with warnings.catch_warnings(record=True) as caught_warnings:
            # ensure no warnings are swallowed
            warnings.simplefilter("always")

            self.assertEqual(
                loader.validated_config({'a': 'value', 'b': 'thing'}),
                {'a': 'value'}
            )

            self.assertEqual(len(caught_warnings), 1)
            self.assertEqual(caught_warnings[0].category, UserWarning)
            self.assertIn("Found unexpected fields", str(caught_warnings[0].message))

    def test_validated_config_returns_unaltered_if_all_keys_present(self):
        loader = PreconfigurationLoader(
            name="foo",
            config_file_directory="bar/baz",
            optional_keys=set(),
            required_keys={'a', 'b'},
        )
        self.assertEqual(
            loader.validated_config({'a': 'value', 'b': 'thing'}),
            {'a': 'value', 'b': 'thing'}
        )

    def test_load_files_raises_if_no_files_present(self):
        with TemporaryDirectory(dir=self.config_package_dir) as temp_dir:
            loader = PreconfigurationLoader(
                name="foo",
                config_file_directory=temp_dir,
                optional_keys=set(),
                required_keys=set()
            )
            with self.assertRaises(FileNotFoundError):
                loader.load_files(['nonexistent.yaml', 'nope.yaml'])

    def test_load_files_returns_only_present_files(self):
        with TemporaryDirectory(dir=self.config_package_dir) as temp_dir:
            loader = PreconfigurationLoader(
                name="foo",
                config_file_directory=temp_dir,
                optional_keys=set(),
                required_keys={'x'},
            )
            with open(os.path.join(temp_dir, 'thing.yaml'), 'w') as existing_yaml_io:
                yaml.dump({'x': 'y'}, existing_yaml_io)
            files = loader.load_files(['not_there.yaml', 'thing.yaml'])
            self.assertEqual(len(files), 1)
            self.assertEqual(files[0], {'x': 'y'})

    def test_load_for_mode_gives_precedent_to_mode_specific(self):
        with TemporaryDirectory(dir=self.config_package_dir) as temp_dir:
            loader = PreconfigurationLoader(
                name="foo",
                config_file_directory=temp_dir,
                optional_keys=set(),
                required_keys={'x', 'y'}
            )
            with open(os.path.join(temp_dir, 'global.yaml'), 'w') as existing_yaml_io:
                yaml.dump({'x': 'y', 'y': 'z'}, existing_yaml_io)

            with open(os.path.join(temp_dir, 'some_mode.yaml'), 'w') as existing_yaml_io:
                yaml.dump({'x': 'abc'}, existing_yaml_io)

            loaded_info = loader.load_for_mode('some_mode')
            self.assertEqual(loaded_info, {'x': 'abc', 'y': 'z'})

    def test_load_for_mode_works_with_global_only(self):
        with TemporaryDirectory(dir=self.config_package_dir) as temp_dir:
            loader = PreconfigurationLoader(
                name="foo",
                config_file_directory=temp_dir,
                optional_keys=set(),
                required_keys={'x', 'y'}
            )
            with open(os.path.join(temp_dir, 'global.yaml'), 'w') as existing_yaml_io:
                yaml.dump({'x': 'y', 'y': 'z'}, existing_yaml_io)

            loaded_info = loader.load_for_mode('some_mode')
            self.assertEqual(loaded_info, {'x': 'y', 'y': 'z'})



from __future__ import annotations  # this lets us annotate functions in class C that return an instance of C

from dataclasses import dataclass
from functools import cached_property
from typing import Any, Iterator, Optional
from typing_extensions import Protocol

from argo.workflows.client import ApiClient as ArgoApiClient,\
    ArchivedWorkflowServiceApi,\
    Configuration as ArgoConfiguration
from argo.workflows.client.models import V1alpha1Workflow, V1alpha1WorkflowList


def generate_argo_archived_workflows_client(host_url: str, access_token: str) -> ArchivedWorkflowServiceApi:
    return ArchivedWorkflowServiceApi(
        api_client=ArgoApiClient(
            configuration=ArgoConfiguration(host=host_url),
            header_name="Authorization",
            header_value=f"Bearer {access_token}"))


# mirrors the call signature of the "fetch list of workflows from the Argo API" call
class ArgoFetchListOperation(Protocol):
    def __call__(
        self,
        *args: Any,
        list_options_continue: Optional[str] = None,
        **kwargs: Any
    ) -> V1alpha1WorkflowList: ...


@dataclass
class ArgoArchivedWorkflowsClient:
    argo_url: str
    access_token: str

    @cached_property
    def client(self) -> ArchivedWorkflowServiceApi:
        return generate_argo_archived_workflows_client(self.argo_url, self.access_token)

    def list_archived_workflows(self) -> Iterator[V1alpha1Workflow]:
        return self._pull_paginated_results(self.client.list_archived_workflows)

    def get_archived_workflow(self, uid: str) -> V1alpha1Workflow:
        return self.client.get_archived_workflow(uid)

    def _pull_paginated_results(self, api_function: ArgoFetchListOperation) -> Iterator[V1alpha1Workflow]:
        results = api_function()

        for result in results.items:
            yield result

        while results.metadata._continue:
            results = api_function(list_options_continue=results.metadata._continue)

            for result in results.items:
                yield result


class ExtendedArgoWorkflow:
    def __init__(self, workflow: V1alpha1Workflow, argo_url: str, access_token: str):
        self._workflow = workflow
        self._inflated = False
        self.client = ArgoArchivedWorkflowsClient(argo_url, access_token)

    def inflate(self) -> ExtendedArgoWorkflow:
        if not self._inflated:
            self._workflow = self.client.get_archived_workflow(self.metadata.uid)
            self._inflated = True

        return self

    def params_dict(self) -> dict[str, Any]:
        return {
            param.name: param.value
            for param in self.spec.arguments.parameters
        }

    # proxy pattern - any function calls not defined in this class are passed to the wrapped workflow object
    def __getattr__(self, name: str) -> Any:
        return getattr(self._workflow, name)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ExtendedArgoWorkflow) and self._workflow == other._workflow



from dataclasses import dataclass

from dagster import resource, StringSource, Field
from dagster.core.execution.context.init import InitResourceContext
from dagster_utils.contrib.google import get_credentials
from dagster_utils.resources.data_repo.api_configuration import RefreshingAccessTokenConfig
from data_repo_client import ApiClient, RepositoryApi, PolicyMemberRequest, PolicyResponse, PolicyModel, SnapshotModel


def build_client(host: str) -> RepositoryApi:
    creds = get_credentials()
    config = RefreshingAccessTokenConfig(host=host, google_credentials=creds)
    client = ApiClient(configuration=config)
    client.client_side_validation = False

    return RepositoryApi(api_client=client)


@resource({
    "api_url": Field(StringSource)
})
def jade_data_repo_client(init_context: InitResourceContext) -> RepositoryApi:
    return build_client(init_context.resource_config["api_url"])


class NoopDataRepoClient:
    @dataclass
    class NoopResult:
        total: int

    @dataclass
    class FakeJobResponse:
        completed: bool
        id: str
        job_status: str

    def create_snapshot(self, snapshot: dict[str, str]) -> FakeJobResponse:
        return NoopDataRepoClient.FakeJobResponse(True, "abcdef", "succeeded")

    def enumerate_datasets(self) -> NoopResult:
        return NoopDataRepoClient.NoopResult(5)

    def retrieve_job(self, job_id: str) -> FakeJobResponse:
        return NoopDataRepoClient.FakeJobResponse(True, "abcdef", "succeeded")

    def bulk_file_load(self, dataset_id: str, bulk_file_load: dict[str, str]) -> FakeJobResponse:
        return NoopDataRepoClient.FakeJobResponse(True, "abcdef", "succeeded")

    def retrieve_job_result(self, job_id: str) -> dict[str, object]:
        return {
            "id": "fake_object_id",
            "name": "fake_object_name",
            "failedFiles": 0
        }

    def apply_dataset_data_deletion(self, id: str, data_deletion_request: dict[str, object]) -> FakeJobResponse:
        return NoopDataRepoClient.FakeJobResponse(True, "abcdef", "succeeded")

    def ingest_dataset(self, id: str, ingest: dict[str, object]) -> FakeJobResponse:
        return NoopDataRepoClient.FakeJobResponse(True, "abcdef", "succeeded")

    def add_snapshot_policy_member(self, id: str, policy_name: str,
                                   policy_member: PolicyMemberRequest) -> PolicyResponse:
        return PolicyResponse(policies=[PolicyModel(name=policy_name, members=[policy_member])])

    def retrieve_snapshot(self, id: str, include: list[str]) -> SnapshotModel:
        return SnapshotModel(id=id, name="fake")


@resource
def noop_data_repo_client(init_context: InitResourceContext) -> NoopDataRepoClient:
    return NoopDataRepoClient()



from contextlib import contextmanager
import os
import shutil
from typing import Iterator


# works as a temporary directory, but with a fixed name
@contextmanager
def EphemeralNamedDirectory(dirname: str, parent_directory: str) -> Iterator[str]:
    if dirname == '' or parent_directory == '':
        raise ValueError('Must provide a directory name to create!')

    target_path = os.path.join(parent_directory, dirname)
    if os.path.exists(target_path):
        raise FileExistsError(f"Tried to create an ephemeral directory at existing path {target_path}")

    try:
        os.mkdir(target_path)
        yield target_path
    finally:
        shutil.rmtree(target_path)




import data_repo_client
from google.auth.transport.requests import Request
from google.auth.credentials import Credentials


class RefreshingAccessTokenConfig(data_repo_client.Configuration):
    """
    Helper class that hooks access to the access_token property so we can
    intercept and potentially refresh if expired
    """

    def __init__(self, host: str, google_credentials: Credentials) -> None:
        super().__init__(host)
        self._google_credentials = google_credentials

    @property
    def access_token(self) -> str:
        # if creds are expired, attempt to refresh
        if not self._google_credentials.valid:
            self._google_credentials.refresh(Request())
        self._access_token: str = self._google_credentials.token
        return self._access_token

    @access_token.setter
    def access_token(self, value: str) -> None:
        self._access_token = value
